import argparse
import logging
import threading
import time
from collections import deque

from . import image_processor
from . import keyboard_handler
from . import screen_capture
from . import terminal_interface
from . import user_interface as ui
from . import wled_discovery
from .pixel_writer import PixelWriter
import wx

logging.basicConfig(
    filename="debug.log",
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(message)s",
)


def get_arguments():
    default_fps = 30
    # Parse command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--fps", type=int, default=default_fps, help="Target FPS")
    parser.add_argument(
        "--title",
        type=str,
        default=None,
        help="Cast window whose title contains this string",
    )
    parser.add_argument(
        "--search-timeout",
        type=int,
        default=3,
        help="Seconds to wait for WLED instance discovery",
    )
    parser.add_argument(
        "--output-resolution",
        type=str,
        default=None,
        help="Specify the output resolution rather than discovering from WLED",
    )
    parser.add_argument(
        "--live-preview",
        default=False,
        help="Show a live preview of the WLED output  the computer screen",
        action="store_true",
    )
    parser.add_argument(
        "--dxcam",
        default=False,
        action="store_true",
        help="Use DirectX Capture (faster, but only for Windows and primary monitor)",
    )
    parser.add_argument(
        "--host",
        default=None,
        type=str,
        help="Specify the IP address of the WLED instance to cast to",
    )
    args = parser.parse_args()
    return args


def main():
    args = get_arguments()
    if args.dxcam:
        from dxcam_capture import capture_dxcam
        capture_function = capture_dxcam
    else:
        capture_function = None

    if args.host is not None:
        selected_wled_host = args.host
    else:
        # Discover WLED instances
        wled_instances = wled_discovery.discover(args.search_timeout)
        if len(wled_instances) == 0:
            return 1

        # Ask user which WLED instance to cast to
        selected_wled_host = (
            wled_discovery.select_instance(wled_instances)
            if len(wled_instances) > 1
            else wled_instances[0]
        )

    # Determine the shape of the LED pixel matrix from WLED
    led_matrix_shape = wled_discovery.get_matrix_shape(selected_wled_host)
    # get the coordinates to capture, use a list to make it mutable
    window = screen_capture.select_window(title=args.title)

    # get the capture coordinates: dict[left, top, width, height]
    capture_box = screen_capture.get_capture_box(window, led_matrix_shape)

    app = wx.App()
    ui_capture_border = ui.TransparentWindow(
        parent=None,
        title="wledcast",
        size=capture_box.getSize(),
        pos=capture_box.getPosition(),
        capture_box=capture_box,
    )
    ui_capture_border.Show()

    # Initialize the pixel writer
    frame_times = deque(maxlen=10)



    def cast():
        while not stop_event.is_set():
            start = time.time()
            # Capture the selected screen
            rgb_array = screen_capture.capture_window_content(
                capture_box, led_matrix_shape, capture_function
            )
            if rgb_array is None:
                print("**Dropped frame**".ljust(40))
                continue
            if args.live_preview:
                image_processor.show_preview(rgb_array, led_matrix_shape)
            # Update the LED matrix via WLED in real-time
            pixel_writer.update_pixels(rgb_array)
            end = time.time()
            dt = end - start
            # Frame rate limiter
            if dt < 1 / args.fps:
                time.sleep(1 / args.fps - dt)

            frame_times.append(end)
        return 0

    stop_event = threading.Event()
    pixel_writer = PixelWriter(selected_wled_host)
    cast_thread = threading.Thread(target=cast, daemon=True)
    terminal_thread = threading.Thread(
        target=terminal_interface.start_terminal_ui,
        args=(image_processor.config, frame_times, capture_box, stop_event),
        daemon=True,
    )
    keyboard_handler.setup_keybinds(ui_capture_border, capture_box, stop_event)

    try:
        terminal_thread.start()
        cast_thread.start()

        app.MainLoop()
    except Exception as e:
        print(f"Unexpected error: {e}")
    finally:
        stop_event.set()
        print("Shutting down...")
        cast_thread.join()
        terminal_thread.join()


if __name__ == "__main__":
    main()



import keyboard
import threading
import wx

from .model import Box
from .user_interface import get_virtual_desktop_size



def setup_keybinds(
    frame: wx.Frame, capture_box: Box, stop_event: threading.Event
) -> callable:
    max_x, max_y = get_virtual_desktop_size()
    aspect_ratio = frame.GetSize().GetWidth() / frame.GetSize().GetHeight()
    # Movement and resizing speeds
    speed_increment = 3
    min_speed = 2
    max_speed = 50

    # Current speed and timers for movement and resizing
    current_speed = {"move": min_speed, "resize": min_speed}

    def adjust_position(delta_x: int, delta_y: int):
        x, y = frame.GetPosition()
        width, height = frame.GetSize()
        new_x = max(0, min(max_x - width, x + delta_x))
        new_y = max(0, min(max_y - height, y + delta_y))
        capture_box.left += new_x - x
        capture_box.top += new_y - y
        wx.CallAfter(frame.SetPosition, (new_x, new_y))

    def adjust_size(delta_w, delta_h):
        width, height = frame.GetSize()
        if delta_w != 0:
            new_width = max(10, min(max_x, width + delta_w))
            new_height = max(10, min(max_y, int(new_width / aspect_ratio)))
        else:
            new_height = max(10, min(max_x, height + delta_h))
            new_width = max(10, min(max_y, int(new_height * aspect_ratio)))
        capture_box.width += new_width - width
        capture_box.height += new_height - height
        wx.CallAfter(frame.SetSize, (new_width, new_height))

    def perform_action(action_type, key):
        # Determine the action based on the key and whether it's a move or resize action
        delta = current_speed[action_type]

        if action_type == "move":
            if key == "left":
                adjust_position(-delta, 0)
            elif key == "right":
                adjust_position(delta, 0)
            elif key == "up":
                adjust_position(0, -delta)
            elif key == "down":
                adjust_position(0, delta)
        elif action_type == "resize":
            if key == "left":
                adjust_size(-delta, 0)
            elif key == "right":
                adjust_size(delta, 0)
            elif key == "up":
                adjust_size(0, -delta)
            elif key == "down":
                adjust_size(0, delta)

    def update_speed(action_type):
        # Accelerate the movement or resizing speed
        if current_speed[action_type] < max_speed:
            current_speed[action_type] += speed_increment

    def on_key_event(event):
        key = event.name
        if key in ["left", "right", "up", "down"]:
            ctrl = keyboard.is_pressed("ctrl")
            alt = keyboard.is_pressed("alt")

            if event.event_type == "down":
                action_type = "move" if ctrl else "resize" if alt else None
                if action_type:
                    perform_action(action_type, key)
                    update_speed(action_type)
            elif event.event_type == "up":
                # Stop the action and reset speed for both move and resize
                for action_type in ["move", "resize"]:
                    current_speed[action_type] = (
                        min_speed if action_type == "move" else min_speed
                    )

        elif key == "esc" and event.event_type == "down":
            stop_event.set()
            wx.CallAfter(frame.Close)

    keyboard.hook(on_key_event)

    return on_key_event


from typing import Union

from . import image_processor
from .model import Box

import mss
import numpy as np
import pywinctl
from PIL import Image


def select_window(title=None) -> pywinctl.Window:
    # Get the list of open windows
    windows: list[pywinctl.Window] = pywinctl.getAllWindows()
    windows = [
        window
        for window in windows
        if not window.isMinimized and not window._parent and window.title
    ]

    if title:
        windows = [window for window in windows if title in window.title]
    if len(windows) == 1:
        return windows[0]
    # Ask the user to select a window from a list numbered from 1 upwards
    for i, window in enumerate(windows):
        print(f"{i + 1}: {window.title}")
    choice = int(input("Enter the number of the window to capture: ")) - 1
    if choice < 0 or choice >= len(windows):
        print("Invalid selection. Please try again.")
        return select_window()

    return windows[choice]


def get_capture_box(window: pywinctl.Window, target_resolution) -> Box:
    # Get the client rectangle of the window
    client_box = Box(
        **{
            "left": window.box.left,
            "top": window.box.top,
            "width": window.box.width,
            "height": window.box.height,
        }
    )

    # crop the largest area in client_rect that matches the aspect ratio of resolution
    target_aspect_ratio = target_resolution[0] / target_resolution[1]

    # Calculate the aspect ratio of the client rectangle
    client_aspect_ratio = client_box.width / client_box.height

    # Check if the aspect ratio of the client rectangle matches the aspect ratio of the resolution
    if client_aspect_ratio > target_aspect_ratio:
        # Crop the width of the client rectangle to match the aspect ratio of the resolution
        new_width = int(client_box.height * target_aspect_ratio)
        center_x = client_box.left + (client_box.width // 2)
        client_box.left = center_x - (new_width // 2)
        client_box.width = new_width
    else:
        # Crop the height of the client rectangle to match the aspect ratio of the resolution
        new_height = int(client_box.width / target_aspect_ratio)
        center_y = client_box.top + (client_box.height // 2)
        client_box.top = center_y - (new_height // 2)
        client_box.height = new_height

    return client_box


def capture_window_content(
    window_box: Box, resolution: tuple[int, int], capture_function: Union[callable, None] = None
) -> np.ndarray | None:
    try:
        if capture_function is not None:
            return capture_function(window_box, resolution)
        # noinspection PyTypeChecker
        return capture_mss(window_box, resolution)
    except Exception as e:
        print(f"\nAn error occurred: {e}")
        return None


def capture_mss(window_box: Box, resolution):
    # Use mss to capture the image
    with mss.mss() as sct:
        sct_img = sct.grab(vars(window_box))
        image = Image.frombytes("RGB", sct_img.size, sct_img.bgra, "raw", "BGRX")
        image = image.resize(resolution, Image.Resampling.BOX)
        # image = image_processor.apply_filters_pillow(image)
        rgb_array = np.array(image)
        rgb_array = image_processor.apply_filters_cv2(
            rgb_array
        )  # apply filters after resizing
        return rgb_array



import json
import os.path
import threading
import time
from collections import deque

from asciimatics.event import KeyboardEvent
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.widgets import Button, Frame, Label, Layout, Text

from .model import Box

def config_editor(
    screen,
    config,
    frame_times: deque,
    capture_box: Box,
    stop_event: threading.Event,
):
    def update_form():
        # Update form values from config
        for key, value in config.items():
            form_data[key].value = str(value)

    def save_config():
        # Save form values to config
        try:
            for key, value in config.items():
                config[key] = float(form_data[key].value)
            with open(os.path.join(os.path.dirname(__file__), "filter.json"), "w") as f:
                f.write(json.dumps(config))
        except ValueError as exc:
            pass  # Handle invalid float conversion

    frame = Frame(
        screen,
        int(screen.height * 2 // 3),
        int(screen.width * 2 // 3),
        title="Edit Configuration",
    )
    frame.set_theme("green")
    layout = Layout([1], fill_frame=True)
    frame.add_layout(layout)

    layout.add_widget(
        Label("Use Ctrl + arrow keys or left mouse button move the capture area")
    )
    layout.add_widget(
        Label("Use Alt + arrow keys or right mouse button to resize the capture area")
    )
    layout.add_widget(Label("Esc to exit"))

    # print FPS: 10 / time for last 10 frames
    fps_label = Label(
        f"Casting {capture_box.width}x{capture_box.height} ({capture_box.left}, {capture_box.top}) to ({capture_box.left + capture_box.width}, {capture_box.top + capture_box.height}) at ~~~fps"
    )
    layout.add_widget(fps_label)

    # Create form fields
    form_data = {
        "sharpen": Text("Sharpen:", "sharpen"),
        "saturation": Text("Saturation:", "saturation"),
        "brightness": Text("Brightness:", "brightness"),
        "contrast": Text("Contrast:", "contrast"),
        "balance_r": Text("Balance R:", "balance_r"),
        "balance_g": Text("Balance G:", "balance_g"),
        "balance_b": Text("Balance B:", "balance_b"),
    }
    for name, field in form_data.items():
        layout.add_widget(field)
    layout.add_widget(Button("Save", save_config))

    frame.fix()
    update_form()

    # Create a scene with the frame
    scenes = [Scene([frame], -1)]
    screen.set_scenes(scenes)

    while not stop_event.is_set():
        if len(frame_times) >= 10:
            fps_label.text = f"Casting {capture_box.width}x{capture_box.height} ({capture_box.left}, {capture_box.top}) to ({capture_box.left + capture_box.width}, {capture_box.top + capture_box.height}) at {round(len(frame_times) / (frame_times[len(frame_times) - 1] - frame_times[0]), 1) if len(frame_times) > 0 else '~~'}fps.)"
        screen.draw_next_frame(repeat=False)
        event = screen.get_event()
        if isinstance(event, KeyboardEvent):
            frame.process_event(event)
        time.sleep(0.05)


def start_terminal_ui(
    config: dict,
    frame_times: deque,
    capture_box: Box,
    stop_event: threading.Event,
):
    def run(screen: Screen):
        config_editor(screen, config, frame_times, capture_box, stop_event)

    Screen.wrapper(run)
    return 0